package br.ime.usp.mac5855.controller;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import br.ime.usp.mac5855.view.ErrorHandler;

public class BeanPopulator {

	private final Map<Class<?>, Converter<?>> converters = new HashMap<Class<?>, Converter<?>>();
	
	public BeanPopulator() {
		registerConverters();
	}
	
	public Object populateObjectWithParams(Class<?> clazz, HttpServletRequest request) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException, InstantiationException, SecurityException, NoSuchMethodException {
		return populateObjectWithParams(clazz, lowerName(clazz.getSimpleName()), request);
	}

	private Object populateObjectWithParams(Class<?> clazz, String basename, HttpServletRequest request) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException, InstantiationException, SecurityException, NoSuchMethodException {
		Object object = clazz.newInstance();
		
		for (Method method : object.getClass().getMethods()) {
			String mname = method.getName();
			if (mname.startsWith("set")) {
				String paramName = getParamName(basename, mname);
				String paramValue = request.getParameter(paramName);
				Class<?> paramClass = method.getParameterTypes()[0];
				
				Object methodParam = null;
				if (paramValue != null && !paramValue.equals("")) {
					Converter<?> converter = converters.get(paramClass);
					if (converter != null) {
						methodParam = converter.convert(paramValue);
						if (methodParam == null) {
							ErrorHandler.add("Conversion error for parameter " + paramName, request);
						}
					} else if (paramClass.isEnum()) {
						methodParam = paramClass.getMethod("valueOf", String.class).invoke(null, paramValue);
					} else {
						throw new ControllerException("Can't populate bean with " + paramClass.getName());
					}
				} else {
					String newBasename = basename + "." + propertyName(method.getName());
					if (hasParameterStartingWith(newBasename, request)) {
						methodParam = populateObjectWithParams(paramClass, newBasename, request);
					}
				}
				
				if (methodParam != null) {
					method.invoke(object, methodParam);
				}
			}
		}
		
		return object;
	}

	private void registerConverters() {
		converters.put(String.class, new Converter<String>() {
			public String convert(String s) {
				return s;
			}
		});
		
		// wrappers
		converters.put(Long.class, new Converter<Long>() {
			public Long convert(String s) {
				try {
					return Long.valueOf(s);
				} catch (NumberFormatException e) {
					return null;
				}
			}
		});
		converters.put(long.class, converters.get(Long.class));
		converters.put(Double.class, new Converter<Double>() {
			public Double convert(String s) {
				try {
					return Double.valueOf(s);
				} catch (NumberFormatException e) {
					return null;
				}
			}
		});
		converters.put(double.class, converters.get(Double.class));
		converters.put(Integer.class, new Converter<Integer>() {
			public Integer convert(String s) {
				try {
					return Integer.valueOf(s);
				} catch (NumberFormatException e) {
					return null;
				}
			}
		});
		converters.put(int.class, converters.get(Integer.class));
		
		// dates
		converters.put(Date.class, new Converter<Date>() {
			public Date convert(String s) {
				try {
					if (s.contains(":")) {
						return new SimpleDateFormat("dd/MM/yyyy HH:mm").parse(s);
					} else {
						return new SimpleDateFormat("dd/MM/yyyy").parse(s);
					}
				} catch (ParseException e) {
					throw new IllegalArgumentException("invalid date pattern: should be dd/MM/yyyy or dd/MM/yyyy HH:mm");
				}
			}
		});
		converters.put(Calendar.class, new Converter<Calendar>() {
			public Calendar convert(String s) {
				 Date date = (Date) converters.get(Date.class).convert(s);
				 Calendar calendar = Calendar.getInstance();
				 calendar.setTime(date);
				 return calendar;
			}
		});
	}
	
	interface Converter<T> { T convert(String s); }
	
	/* string utils methods */
	private boolean hasParameterStartingWith(String basename, HttpServletRequest request) {
		@SuppressWarnings("unchecked")
		Enumeration<String> enumeration = request.getParameterNames();
		while(enumeration.hasMoreElements()) {
			String paramName = enumeration.nextElement();
			if (paramName.startsWith(basename+".")) {
				return true;
			}
		}
		return false;
	}
	
	private String getParamName(String basename, String acessorName) {
		return basename + "." + propertyName(acessorName); 
	}
	
	private String lowerName(String name) {
		return 	Character.toLowerCase(name.charAt(0)) + name.substring(1);
	}
	
	private String propertyName(String acessorName) {
		return lowerName(acessorName.substring(3));
	}
}
