package br.ime.usp.mac5855.controller;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import br.ime.usp.mac5855.view.ErrorHandler;

/**
 * The controller servlet that takes requests, find the corresponding controller class/method
 * and invoke that for us. Also does the view redirection after action invocation.
 * 
 * @author sergio
 */
public class ControllerServlet extends HttpServlet {
	private static final long serialVersionUID = 1L;
	
	private Map<String, Class<?>> controllers;
	private String jspLocation;

	private ServletContext servletContext;
	
	@Override
	@SuppressWarnings("unchecked")
	public void init(ServletConfig config) throws ServletException {
		super.init(config);
		
		controllers = (Map<String, Class<?>>) config.getServletContext().getAttribute("mappedControllers");
		jspLocation = config.getServletContext().getInitParameter("jsp-location");
		servletContext = config.getServletContext();
	}
	
	@Override
	protected void service(HttpServletRequest request, HttpServletResponse response) 
		throws ServletException, IOException {

		String requested = getRequestedAction(request);
		String[] names = requested.split("\\.");
		
		Class<?> controllerClass = loadControllerClass(names[0]);
		
		String result;
		if (hasCredentials(controllerClass, request.getSession())) {
			Method method = findRequestedControllerMethod(controllerClass, names[1]);
			result = (String) executeRequestedMethod(controllerClass, method, request, response);
		} else {
			ErrorHandler.add("É preciso estar logado para isso", request);
			result = getLoginPage(controllerClass);
		}
		
		// default result
		if (result == null) {
			result = getDefaultView(names[0], names[1]);
		}
		
		redirectToView(result, request, response);	
	}

	private boolean hasCredentials(Class<?> controllerClass, HttpSession session) {
		Controller controller = controllerClass.getAnnotation(Controller.class);
		if (controller.shouldBeLoggedInAs().equals("")) return true;
		if (session.getAttribute(controller.shouldBeLoggedInAs()) == null) return false;
		return true;
	}

	private String getLoginPage(Class<?> controllerClass) {
		Controller controller = controllerClass.getAnnotation(Controller.class);
		return "redirect:" + servletContext.getInitParameter("login-" + controller.shouldBeLoggedInAs());
	}

	/**
	 * The action execution. It takes a method, creates an instance of that controller class
	 * and invokes the method. The method can have no parameter or can receive request and/or 
	 * response. Returns the method invocation result.
	 * 
	 * @param controllerClass the controller class 
	 * @param method the action method from a controller class
	 * @param request The request
	 * @param response The response
	 * @return the method invocation result
	 */
	private Object executeRequestedMethod(Class<?> controllerClass, Method method, HttpServletRequest request, HttpServletResponse response) {
			try {
				Object controller = controllerClass.newInstance();
				
				// in order to allow generic inheritance, we need to make a map of generic variables
				Type superclass = controllerClass.getGenericSuperclass();
				Map<String,Class<?>> genericParams = new HashMap<String, Class<?>>();
				
				if (superclass != null && superclass instanceof ParameterizedType) {
					ParameterizedType parameterizedType = (ParameterizedType) superclass;
					
					TypeVariable<?>[] typeParams = controllerClass.getSuperclass().getTypeParameters();
					for (int i = 0; i < typeParams.length; i++) {
						Class<?> actualType = (Class<?>) parameterizedType.getActualTypeArguments()[i];
						genericParams.put(typeParams[i].getName(), actualType);
					}
				}

				// deal with request / response parameter injection and
				// also simple beans populated from request parameters
				Type[] parameterTypes = method.getGenericParameterTypes();
				Object[] params = new Object[parameterTypes.length];
				
				for (int i = 0; i < parameterTypes.length; i++) {
					if (parameterTypes[i] instanceof Class) {
						// regular types
						Class<?> clazz = (Class<?>) parameterTypes[i];
						if (clazz.equals(HttpServletRequest.class)) {
							params[i] = request;
						} else if (clazz.equals(HttpServletResponse.class)) {
							params[i] = response;
						} else if (clazz.equals(Output.class)) {
							params[i] = new Output(request);
						} else {
							params[i] = new BeanPopulator().populateObjectWithParams(clazz, request);
						}
					} else if (parameterTypes[i] instanceof TypeVariable){
						// generic type
						TypeVariable<?> tv = (TypeVariable<?>) parameterTypes[i];
						Class<?> clazz = genericParams.get(tv.getName());
						params[i] = clazz.newInstance();
						params[i] = new BeanPopulator().populateObjectWithParams(clazz, request);
					}
				}
				
				// invoke method
				return method.invoke(controller, params);
				
			} catch (InstantiationException e) {
				throw new ControllerException("Could not instantiate class " 
						+ method.getDeclaringClass().getName(), e);
			} catch (IllegalAccessException e) {
				throw new ControllerException(e.getMessage());
			} catch (IllegalArgumentException e) {
				throw new ControllerException("Could not invoke method " + method.getName() 
						+ " ("+ method.getDeclaringClass().getName() +")", e);
			} catch (InvocationTargetException e) {
				throw new ControllerException(e.getCause());			
			} catch (SecurityException e) {
				throw new ControllerException(e.getCause());
			} catch (NoSuchMethodException e) {
				throw new ControllerException(e.getCause());			
			}
	}

	private Method findRequestedControllerMethod(Class<?> controllerClass, String methodName) {
		Method controllerMethod = searchRequestedMethod(controllerClass.getDeclaredMethods(), methodName);
		
		// if not found, search all methods, including superclass'
		if (controllerMethod == null) {
			controllerMethod = searchRequestedMethod(controllerClass.getMethods(), methodName);
		}
		
		// if not found
		if (controllerMethod == null) {
			throw new ControllerException("Method not found in controller: "
					+ methodName + " (" + controllerClass.getName() + ")");
		}
		
		return controllerMethod;
	}

	private Method searchRequestedMethod(Method[] methods, String methodName) {
		Method controllerMethod = null;
		
		for (Method method : methods) {
			if (method.isBridge())
				continue; // ignores bridge methods for erasure
			
			if (method.getName().equals(methodName)) {
				if (controllerMethod == null) {
					controllerMethod = method;
				} else {
					throw new ControllerException("Cannot have two methods with the same name in controller: " + methodName);
				}
			}
		}
		return controllerMethod;
	}

	private Class<?> loadControllerClass(String controllerName) {
		Class<?> controllerClass = controllers.get(controllerName);
		
		if (controllerClass == null) {
			throw new ControllerException("Could not found controller named " + controllerName);
		}
		return controllerClass;
	}

	/**
	 * Given a request to /context/controller.method.action, returns controller.method 
	 */
	private String getRequestedAction(HttpServletRequest request) {
		String contextPath = request.getContextPath();
		String uri = request.getRequestURI().substring(contextPath.length() + 1);
		
		if (!uri.matches("([a-zA-Z0-9]+\\.){2}action")) {
			throw new ControllerException("Wrong URI format: " + uri);
		}

		return uri.substring(0, uri.lastIndexOf('.'));
	}

	/**
	 * Helper method that simply returns the default view JSP file if no result is found
	 * after the method invocation. 
	 */
	private String getDefaultView(String controller, String method) {
		return controller+ "/"+ method + ".jsp";
	}

	/**
	 * Helper method that does a server or client redirection based on the result string
	 * 
	 * @param result The result string
	 * @param request The request
	 * @param response The response
	 */
	private void redirectToView(String result, HttpServletRequest request, HttpServletResponse response)
			throws IOException, ServletException {
		if (result.startsWith("redirect:")) {
			response.sendRedirect(result.substring("redirect:".length()));
		} else if (!result.equals("")) {
			request.getRequestDispatcher(jspLocation + "/" + result).forward(request, response);
		}
	}
}