package br.ime.usp.mac5855.controller;

import java.io.File;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.servlet.ServletContext;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;

/**
 * Scans WEB-INF/classes looking for classes with @Controller and
 * register everything in an application context attribute called
 * mappedControllers.
 * 
 * @author sergio
 */
public class RegisterControllersAtStartup implements ServletContextListener{

	private String webInfClasses;

	/**
	 * On context initialization, we scan the classpath and register
	 * all components. 
	 */
	public void contextInitialized(ServletContextEvent event) {
		ServletContext servletContext = event.getServletContext();
		
		webInfClasses = servletContext.getRealPath("/WEB-INF/classes");
		Map<String, Class<?>> controllers = new HashMap<String, Class<?>>();

		registerControllersAt(new File(webInfClasses), controllers);
		
		// wraps as unmodifiable and put in servlet context
		servletContext.setAttribute("mappedControllers", Collections.unmodifiableMap(controllers));
	}

	/**
	 * Helper method to scan directories recursively and find all .class files
	 * @param directory A directory to be scanned
	 * @param controllers Map to put all found controllers
	 */
	private void registerControllersAt(File directory, Map<String, Class<?>> controllers) {
		for (File file: directory.listFiles()){
			if (file.isDirectory()) {
				registerControllersAt(file, controllers);
			} else if (file.getName().endsWith(".class")){
				String classFileName = file.getPath().substring(webInfClasses.length());
				processClassFile(classFileName, controllers);
			}
		}
	}

	/**
	 * Helper method to analyze a .class file and determine if it's a @Controller
	 * @param classFileName A file name belonging to a java .class
	 * @param controllers Map to put all found controllers
	 */
	private void processClassFile(String classFileName, Map<String, Class<?>> controllers) {
		Class<?> clazz = loadClassByClassFile(classFileName);
		if (clazz.isAnnotationPresent(Controller.class)) {
			Controller controller = clazz.getAnnotation(Controller.class);
			String controllerName = controller.name();

			// checks if component name is sintactaly correct
			if (!controllerName.matches("[a-zA-Z0-9]+")) {
				throw new ControllerException("Controller name not allowed: " 
						+ controllerName + " ("+ clazz.getName() +")");
			}
			
			// checks if this controller is already registered
			if (controllers.containsKey(controllerName)) {
				throw new IllegalArgumentException("Could not register " + clazz.getName() 
						+ ": Already have a controller with name " + controllerName 
						+ " (" + controllers.get(controllerName).getName() 	+ ")");
			}
			
			// register controller
			controllers.put(controllerName, clazz);
		}
	}

	/**
	 * Given a file name, return a Class object
	 * @param classFileName the file name ending in .class
	 * @return the loaded class
	 */
	private Class<?> loadClassByClassFile(String classFileName) {
		String className = classFileName
			.substring(0, classFileName.length() - 6)
			.substring(1)
			.replace('/', '.')
			.replace('\\','.');

		try {
			return Class.forName(className);
		} catch (ClassNotFoundException e) {
			throw new ControllerException(e);
		}
	}

	// nothing to be done on context destroy
	public void contextDestroyed(ServletContextEvent arg0) {
	}

}
